
/********************************************
field.c
copyright 1991, Michael D. Brennan

This is a source file for mawk, an implementation of
the AWK programming language.

Mawk is distributed without warranty under the terms of
the GNU General Public License, version 2, 1991.
********************************************/

/* $Log: field.c,v $
 * Revision 1.5  1995/06/18  19:17:47  mike
 * Create a type Int which on most machines is an int, but on machines
 * with 16bit ints, i.e., the PC is a long.  This fixes implicit assumption
 * that int==long.
 *
 * Revision 1.4  1994/10/08  19:15:38  mike
 * remove SM_DOS
 *
 * Revision 1.3  1993/07/14  12:32:39  mike
 * run thru indent
 *
 * Revision 1.2	 1993/07/14  12:22:11  mike
 * rm SIZE_T and (void) casts
 *
 * Revision 1.1.1.1  1993/07/03	 18:58:12  mike
 * move source to cvs
 *
 * Revision 5.7	 1993/05/08  18:06:00  mike
 * mawk_null_split
 *
 * Revision 5.6	 1993/02/13  21:57:25  mike
 * merge patch3
 *
 * Revision 5.5	 1993/01/01  21:30:48  mike
 * split mawk_new_STRING() into mawk_new_STRING and mawk_new_STRING0
 *
 * Revision 5.4.1.2  1993/01/20	 12:53:08  mike
 * d_to_l()
 *
 * Revision 5.4.1.1  1993/01/15	 03:33:42  mike
 * patch3: safer double to int conversion
 *
 * Revision 5.4	 1992/11/29  22:52:11  mike
 * double->string conversions uses long ints for 16/32 bit
 * compatibility.
 * Fixed small LM_DOS mawk_bozo.
 *
 * Revision 5.3	 1992/08/17  14:21:10  brennan
 * patch2: After parsing, only bi_sprintf() uses string_buff.
 *
 * Revision 5.2	 1992/07/10  16:17:10  brennan
 * MsDOS: remove NO_BINMODE macro
 *
 * Revision 5.1	 1991/12/05  07:55:57  brennan
 * 1.1 pre-release
 *
*/


/* field.c */

#include "mawk.h"
#include "field.h"
#include "init.h"
#include "memory.h"
#include "scan.h"
#include "bi_vars.h"
#include "repl.h"
#include "regexp.h"

static void PROTO(build_field0, (mawk_state_t *)) ;
static void PROTO(set_rs_shadow, (mawk_state_t *)) ;
static void PROTO(load_pfield, (mawk_state_t *, char *, CELL *)) ;
static void PROTO(load_field_ov, (mawk_state_t *)) ;

static void
set_rs_shadow(mawk_state_t *MAWK)
{
   CELL c ;
   STRING *sval ;
   char *s ;
   unsigned len ;

   if (MAWK->posix_space_flag && MAWK->mawk_state == EXECUTION)
      MAWK->scan_code['\n'] = SC_UNEXPECTED ;

   if (MAWK->rs_shadow.type == SEP_STR)
      free_STRING((STRING *) MAWK->rs_shadow.ptr) ;

   mawk_cast_for_split(MAWK, mawk_cellcpy(MAWK, &c, RS)) ;
   switch (c.type)
   {
      case C_RE:
	 if (s = mawk_is_string_split(c.ptr, &len))
	 {
	    if (len == 1)
	    {
	       MAWK->rs_shadow.type = SEP_CHAR ;
	       MAWK->rs_shadow.c = s[0] ;
	    }
	    else
	    {
	       MAWK->rs_shadow.type = SEP_STR ;
	       MAWK->rs_shadow.ptr = (PTR) mawk_new_STRING(MAWK, s) ;
	    }
	 }
	 else
	 {
	    MAWK->rs_shadow.type = SEP_RE ;
	    MAWK->rs_shadow.ptr = c.ptr ;
	 }
	 break ;

      case C_SPACE:
	 MAWK->rs_shadow.type = SEP_CHAR ;
	 MAWK->rs_shadow.c = ' ' ;
	 break ;

      case C_SNULL:		/* RS becomes one or more blank lines */
	 if (MAWK->mawk_state == EXECUTION)  MAWK->scan_code['\n'] = SC_SPACE ;
	 MAWK->rs_shadow.type = SEP_MLR ;
	 sval = mawk_new_STRING(MAWK, "\n\n+") ;
	 MAWK->rs_shadow.ptr = mawk_re_compile(MAWK, sval) ;
	 free_STRING(sval) ;
	 break ;

      default:
	 mawk_bozo(MAWK, "bad cell in set_rs_shadow") ;
   }
}

static void
load_pfield(MAWK, name, cp)
   mawk_state_t *MAWK;
   char *name ;
   CELL *cp ;
{
   SYMTAB *stp ;

   stp = mawk_insert(MAWK, name) ; stp->type = ST_FIELD ;
   stp->stval.cp = cp ;
}

/* mawk_initialize $0 and the pseudo fields */
void
mawk_field_init(mawk_state_t *MAWK)
{
   MAWK->field[0].type = C_STRING ;
   MAWK->field[0].ptr = (PTR) & MAWK->null_str ;
   MAWK->null_str.ref_cnt++ ;

   load_pfield(MAWK, "NF", NF) ;
   NF->type = C_DOUBLE ;
   NF->dval = 0.0 ;

   load_pfield(MAWK, "RS", RS) ;
   RS->type = C_STRING ;
   RS->ptr = (PTR) mawk_new_STRING(MAWK, "\n") ;
   /* rs_shadow already set */

   load_pfield(MAWK, "FS", FS) ;
   FS->type = C_STRING ;
   FS->ptr = (PTR) mawk_new_STRING(MAWK, " ") ;
   /* fs_shadow is already set */

   load_pfield(MAWK, "OFMT", OFMT) ;
   OFMT->type = C_STRING ;
   OFMT->ptr = (PTR) mawk_new_STRING(MAWK, "%.6g") ;

   load_pfield(MAWK, "CONVFMT", CONVFMT) ;
   CONVFMT->type = C_STRING ;
   CONVFMT->ptr = OFMT->ptr ;
   string(OFMT)->ref_cnt++ ;
}



void
mawk_set_field0(MAWK, s, len)
   mawk_state_t *MAWK;
   char *s ;
   unsigned len ;
{
   cell_destroy(&MAWK->field[0]) ;
   MAWK->nf = -1 ;

   if (len)
   {
      MAWK->field[0].type = C_MBSTRN ;
      MAWK->field[0].ptr = (PTR) mawk_new_STRING0(MAWK, len) ;
      memcpy(string(&MAWK->field[0])->str, s, len) ;
   }
   else
   {
      MAWK->field[0].type = C_STRING ;
      MAWK->field[0].ptr = (PTR) & MAWK->null_str ;
      MAWK->null_str.ref_cnt++ ;
   }
}



/* split field[0] into $1, $2 ... and set NF  */

void
mawk_split_field0(mawk_state_t *MAWK)
{
   register CELL *cp ;
   register int cnt ;
   CELL c ;			 /* copy field[0] here if not string */


   if (MAWK->field[0].type < C_STRING)
   {
      mawk_cast1_to_s(MAWK, mawk_cellcpy(MAWK, &c, MAWK->field + 0)) ;
      cp = &c ;
   }
   else	 cp = &MAWK->field[0] ;

   if (string(cp)->len == 0)  MAWK->nf = 0 ;
   else
   {
      switch (MAWK->fs_shadow.type)
      {
	 case C_SNULL:		/* FS == "" */
	    MAWK->nf = mawk_null_split(MAWK, string(cp)->str) ;
	    break ;

	 case C_SPACE:
	    MAWK->nf = mawk_space_split(MAWK, string(cp)->str, string(cp)->len) ;
	    break ;

	 default:
	    MAWK->nf = mawk_re_split(MAWK, string(cp)->str, MAWK->fs_shadow.ptr) ;
	    break ;
      }

   }

   cell_destroy(NF) ;
   NF->type = C_DOUBLE ;
   NF->dval = (double) MAWK->nf ;

   if (MAWK->nf > MAX_SPLIT)
   {
      cnt = MAX_SPLIT ; load_field_ov(MAWK) ;
   }
   else	 cnt = MAWK->nf ;

   while (cnt > 0)
   {
      cell_destroy(MAWK->field + cnt) ;
      MAWK->field[cnt].ptr = (PTR) split_buff[cnt - 1] ;
      MAWK->field[cnt--].type = C_MBSTRN ;
   }

   if (cp == &c)  free_STRING(string(cp)) ;
}

/*
  assign CELL *cp to field or pseudo field
  and take care of all side effects
*/

void
mawk_field_assign(MAWK, fp, cp)
   mawk_state_t *MAWK;
   register CELL *fp ;
   CELL *cp ;
{
   CELL c ;
   int i, j ;

   /* the most common case first */
   if (fp == MAWK->field)
   {
      cell_destroy(MAWK->field) ;
      mawk_cellcpy(MAWK, fp, cp) ;
      MAWK->nf = -1 ;
      return ;
   }

   /* its not important to do any of this fast */

   if (MAWK->nf < 0)	mawk_split_field0(MAWK) ;

#ifdef  MSDOS
   if (!SAMESEG(fp, MAWK->field))
   {
      i = -1 ;
      goto lm_dos_label ;
   }
#endif

   switch (i = (fp - MAWK->field))
   {

      case NF_field:

	 cell_destroy(NF) ;
	 mawk_cellcpy(MAWK, NF, mawk_cellcpy(MAWK, &c, cp)) ;
	 if (c.type != C_DOUBLE)  mawk_cast1_to_d(MAWK, &c) ;

	 if ((j = d_to_i(c.dval)) < 0)
	    mawk_rt_error(MAWK, "negative value assigned to NF") ;

	 if (j > MAWK->nf)
	    for (i = MAWK->nf + 1; i <= j; i++)
	    {
	       cp = field_ptr(i) ;
	       cell_destroy(cp) ;
	       cp->type = C_STRING ;
	       cp->ptr = (PTR) & MAWK->null_str ;
	       MAWK->null_str.ref_cnt++ ;
	    }

	 MAWK->nf = j ;
	 build_field0(MAWK) ;
	 break ;

      case RS_field:
	 cell_destroy(RS) ;
	 mawk_cellcpy(MAWK, RS, cp) ;
	 set_rs_shadow(MAWK) ;
	 break ;

      case FS_field:
	 cell_destroy(FS) ;
	 mawk_cast_for_split(MAWK, mawk_cellcpy(MAWK, &MAWK->fs_shadow, mawk_cellcpy(MAWK, FS, cp))) ;
	 break ;

      case OFMT_field:
      case CONVFMT_field:
	 /* If the user does something stupid with OFMT or CONVFMT,
	   we could crash.
	   We'll make an attempt to protect ourselves here.  This is
	   why OFMT and CONVFMT are pseudo fields.

	   The ptrs of OFMT and CONVFMT always have a valid STRING,
	   even if assigned a DOUBLE or NOINIT
	*/

	 free_STRING(string(fp)) ;
	 mawk_cellcpy(MAWK, fp, cp) ;
	 if (fp->type < C_STRING)	/* !! */
	    fp->ptr = (PTR) mawk_new_STRING(MAWK, "%.6g") ;
	 else if (fp == CONVFMT)
	 {
	    /* It's a string, but if it's really goofy and CONVFMT,
	     it could still mawk_damage us. Test it .
	  */
	    char xbuff[512] ;

	    xbuff[256] = 0 ;
	    sprintf(xbuff, string(fp)->str, 3.1459) ;
	    if (xbuff[256])
	       mawk_rt_error(MAWK, "CONVFMT assigned unusable value") ;
	 }
	 break ;

#ifdef MSDOS
       lm_dos_label:
#endif

      default:			/* $1 or $2 or ... */


	 cell_destroy(fp) ;
	 mawk_cellcpy(MAWK, fp, cp) ;

	 if (i < 0 || i > MAX_SPLIT)  i = mawk_field_addr_to_index(MAWK, fp) ;

	 if (i > MAWK->nf)
	 {
	    for (j = MAWK->nf + 1; j < i; j++)
	    {
	       cp = field_ptr(j) ;
	       cell_destroy(cp) ;
	       cp->type = C_STRING ;
	       cp->ptr = (PTR) & MAWK->null_str ;
	       MAWK->null_str.ref_cnt++ ;
	    }
	    MAWK->nf = i ;
	    cell_destroy(NF) ;
	    NF->type = C_DOUBLE ;
	    NF->dval = (double) i ;
	 }

	 build_field0(MAWK) ;

   }
}


/* construct field[0] from the other fields */

static void
build_field0(mawk_state_t *MAWK)
{


#ifdef DEBUG
   if (MAWK->nf < 0)	mawk_bozo(MAWK, "nf <0 in build_field0") ;
#endif

   cell_destroy(MAWK->field + 0) ;

   if (MAWK->nf == 0)
   {
      MAWK->field[0].type = C_STRING ;
      MAWK->field[0].ptr = (PTR) & MAWK->null_str ;
      MAWK->null_str.ref_cnt++ ;
   }
   else if (MAWK->nf == 1)
   {
      mawk_cellcpy(MAWK, MAWK->field, MAWK->field + 1) ;
   }
   else
   {
      CELL c ;
      STRING *ofs, *tail ;
      unsigned len ;
      register CELL *cp ;
      register char *p, *q ;
      int cnt ;
      CELL **fbp, *cp_limit ;


      mawk_cast1_to_s(MAWK, mawk_cellcpy(MAWK, &c, OFS)) ;
      ofs = (STRING *) c.ptr ;
      mawk_cast1_to_s(MAWK, mawk_cellcpy(MAWK, &c, field_ptr(MAWK->nf))) ;
      tail = (STRING *) c.ptr ;
      cnt = MAWK->nf - 1 ;

      len = cnt * ofs->len + tail->len ;

      fbp = MAWK->fbank ; cp_limit = MAWK->field + FBANK_SZ ;
      cp = MAWK->field + 1 ;

      while (cnt-- > 0)
      {
	 if (cp->type < C_STRING)
	 {			/* use the string field temporarily */
	    if (cp->type == C_NOINIT)
	    {
	       cp->ptr = (PTR) & MAWK->null_str ;
	       MAWK->null_str.ref_cnt++ ;
	    }
	    else  /* its a double */
	    {
	       Int ival ;
	       char xbuff[260] ;

	       ival = mawk_d_to_I(cp->dval) ;
	       if (ival == cp->dval)  sprintf(xbuff, INT_FMT, ival) ;
	       else  sprintf(xbuff, string(CONVFMT)->str, cp->dval) ;

	       cp->ptr = (PTR) mawk_new_STRING(MAWK, xbuff) ;
	    }
	 }

	 len += string(cp)->len ;

	 if (++cp == cp_limit)
	 {
	    cp = *++fbp ;
	    cp_limit = cp + FBANK_SZ ;
	 }

      }

      MAWK->field[0].type = C_STRING ;
      MAWK->field[0].ptr = (PTR) mawk_new_STRING0(MAWK, len) ;

      p = string(MAWK->field)->str ;

      /* walk it again , putting things together */
      cnt = MAWK->nf-1 ; fbp = MAWK->fbank ;
      cp = MAWK->field+1 ; cp_limit = MAWK->field + FBANK_SZ ;
      while (cnt-- > 0)
      {
	 memcpy(p, string(cp)->str, string(cp)->len) ;
	 p += string(cp)->len ;
	 /* if not really string, free temp use of ptr */
	 if (cp->type < C_STRING)  free_STRING(string(cp)) ;
	 if (++cp == cp_limit)
	 {
	    cp = *++fbp ;
	    cp_limit = cp + FBANK_SZ ;
	 }
	 /* add the separator */
	 q = ofs->str ;	 while( *q )  *p++ = *q++ ;
      }
      /* tack tail on the end */
      memcpy(p, tail->str, tail->len) ;

      /* cleanup */
      free_STRING(tail) ; free_STRING(ofs) ;
   }
}

/* We are assigning to a CELL and we aren't sure if its
   a field */

void
mawk_slow_cell_assign(MAWK, target, source)
   mawk_state_t *MAWK;
   register CELL *target ;
   CELL *source ;
{
   if (

#ifdef MSDOS		/* the dreaded segment nonsense */
	 SAMESEG(target, MAWK->field) &&
#endif
	 target >= MAWK->field && target <= LAST_PFIELD)
      mawk_field_assign(MAWK, target, source) ;
   else
   {
      CELL **p = MAWK->fbank + 1 ;

      while (*p)
      {
	 if (
#ifdef  MSDOS
	       SAMESEG(target, *p) &&
#endif
	       target >= *p && target < *p + FBANK_SZ)
	 {
	    mawk_field_assign(MAWK, target, source) ;
	    return ;
	 }
	 p++ ;
      }
      /* its not a field */
      cell_destroy(target) ;
      mawk_cellcpy(MAWK, target, source) ;
   }
}

int
mawk_field_addr_to_index(MAWK, cp)
   mawk_state_t *MAWK;
   CELL *cp ;
{
   CELL **p = MAWK->fbank ;

   while (

#ifdef MSDOS
	    !SAMESEG(cp, *p) ||
#endif

	    cp < *p || cp >= *p + FBANK_SZ)
      p++ ;

   return ((p - MAWK->fbank) << FB_SHIFT) + (cp - *p) ;
}

/*------- more than 1 fbank needed  ------------*/

/*
  compute the address of a field with index
  > MAX_SPLIT
*/

CELL *
mawk_slow_field_ptr(MAWK, i)
   mawk_state_t *MAWK;
   register int i ;
{

   if (i > MAWK->max_field)
   {
      int j ;

      if (i > MAX_FIELD)
	 mawk_rt_overflow(MAWK, "maximum number of fields", MAX_FIELD) ;

      j = 1 ;
      while (MAWK->fbank[j])	j++ ;

      do
      {
	 MAWK->fbank[j] = (CELL *) zmalloc(MAWK, sizeof(CELL) * FBANK_SZ) ;
	 memset(MAWK->fbank[j], 0, sizeof(CELL) * FBANK_SZ) ;
	 j++ ;
	 MAWK->max_field += FBANK_SZ ;
      }
      while (i > MAWK->max_field);
   }

   return &MAWK->fbank[i >> FB_SHIFT][i & (FBANK_SZ - 1)] ;
}

/*
  $0 split into more than MAX_SPLIT fields,
  $(MAX_FIELD+1) ... are on the split_ov_list.
  Copy into fields which start at fbank[1]
*/

static void
load_field_ov(mawk_state_t *MAWK)
{
   register SPLIT_OV *p ;	 /* walks split_ov_list */
   register CELL *cp ;		 /* target of copy */
   int j ;			 /* current fbank[] */
   CELL *cp_limit ;		 /* change fbank[] */
   SPLIT_OV *q ;		 /* trails p */

   /* make sure the fields are allocated */
   mawk_slow_field_ptr(MAWK, MAWK->nf) ;

   p = MAWK->split_ov_list ; MAWK->split_ov_list = (SPLIT_OV*) 0 ;
   j = 1 ; cp = MAWK->fbank[j] ; cp_limit = cp + FBANK_SZ ;
   while (p)
   {
      cell_destroy(cp) ;
      cp->type = C_MBSTRN ;
      cp->ptr = (PTR) p->sval ;

      if (++cp == cp_limit)
      {
	 cp = MAWK->fbank[++j] ; cp_limit = cp + FBANK_SZ ;
      }

      q = p ; p = p->link ; ZFREE(q) ;
   }
}


#if  MSDOS

int
mawk_binmode()			/* read current value of BINMODE */
{
   CELL c ;

   mawk_cast1_to_d(MAWK, mawk_cellcpy(MAWK, &c, BINMODE)) ;
   return d_to_i(c.dval) ;
}

/* set BINMODE and RS and ORS
   from environment or -W mawk_binmode=   */

void
set_binmode(x)
   int x ;
{
   CELL c ;

   /* set RS */
   c.type = C_STRING ;
   c.ptr = (PTR) mawk_new_STRING(MAWK, (x & 1) ? "\r\n" : "\n") ;
   mawk_field_assign(RS, &c) ;
   free_STRING(string(&c)) ;

   /* set ORS */
   cell_destroy(ORS) ;
   ORS->type = C_STRING ;
   ORS->ptr = (PTR) mawk_new_STRING(MAWK, (x & 2) ? "\r\n" : "\n") ;

   cell_destroy(BINMODE) ;
   BINMODE->type = C_DOUBLE ;
   BINMODE->dval = (double) x ;
}

#endif /* MSDOS */
