
/********************************************
scan.c
copyright 1991, Michael D. Brennan

This is a source file for mawk, an implementation of
the AWK programming language.

Mawk is distributed without warranty under the terms of
the GNU General Public License, version 2, 1991.
********************************************/


/* $Log: scan.c,v $
 * Revision 1.8  1996/07/28 21:47:05  mike
 * gnuish patch
 *
 * Revision 1.7  1995/06/18  19:42:24  mike
 * Remove some redundant declarations and add some prototypes
 *
 * Revision 1.6  1995/06/10  16:57:52  mike
 * silently exit(0) if no program
 * always add a '\n' on eof in scan_fillbuff()
 *
 * Revision 1.5  1995/06/06  00:18:33  mike
 * change mawk_exit(1) to mawk_exit(2)
 *
 * Revision 1.4  1994/09/23  00:20:04  mike
 * minor bug fix: handle \ in mawk_eat_nl()
 *
 * Revision 1.3  1993/07/17  00:45:21  mike
 * indent
 *
 * Revision 1.2	 1993/07/04  12:52:09  mike
 * start on autoconfig changes
 *
 * Revision 1.1.1.1  1993/07/03	 18:58:20  mike
 * move source to cvs
 *
 * Revision 5.6	 1993/02/13  21:57:33  mike
 * merge patch3
 *
 * Revision 5.5	 1993/01/01  21:30:48  mike
 * split mawk_new_STRING() into mawk_new_STRING and mawk_new_STRING0
 *
 * Revision 5.4.1.1  1993/01/15	 03:33:50  mike
 * patch3: safer double to int conversion
 *
 * Revision 5.4	 1992/11/29  18:57:50  mike
 * field expressions convert to long so 16 bit and 32 bit
 * systems behave the same
 *
 * Revision 5.3	 1992/07/08  15:43:41  brennan
 * patch2: length returns.  I am a wimp
 *
 * Revision 5.2	 1992/02/21  14:16:53  brennan
 * fix:	 getline <=
 *
 * Revision 5.1	 91/12/05  07:56:27  brennan
 * 1.1 pre-release
 *
*/
#include  "mawk.h"
#include  <limits.h>
#define _POSIX_SOURCE
#include  <string.h>
#include  "scan.h"
#include  "memory.h"
#include  "field.h"
#include  "init.h"
#include  "fin.h"
#include  "repl.h"
#include  "code.h"
#include  "bi_vars.h"

#ifndef PATH_MAX
#define PATH_MAX 1024
#endif

#ifndef	  NO_FCNTL_H
#include  <fcntl.h>
#endif

#include  "files.h"


/* static functions */
static void PROTO(scan_fillbuff, (mawk_state_t *MAWK)) ;
int PROTO(mawk_scan_open, (mawk_state_t *MAWK)) ;
static int PROTO(slow_next, (mawk_state_t *MAWK)) ;
static void PROTO(eat_comment, (mawk_state_t *MAWK)) ;
static void PROTO(eat_semi_colon, (mawk_state_t *MAWK)) ;
static double PROTO(collect_decimal, (mawk_state_t *, int, int *)) ;
static int PROTO(collect_string, (mawk_state_t *)) ;
static int PROTO(collect_RE, (mawk_state_t *)) ;


/*-----------------------------
  program file management
 *----------------------------*/

void
mawk_scan_init(MAWK, cmdline_program)
   mawk_state_t *MAWK;
   char *cmdline_program ;
{
   if (cmdline_program)
   {
      MAWK->ps.program_fd = -1 ;		 /* command line program */
      MAWK->ps.program_string = mawk_new_STRING0(MAWK, strlen(cmdline_program) + 1) ;
      strcpy(MAWK->ps.program_string->str, cmdline_program) ;
      /* simulate file termination */
      MAWK->ps.program_string->str[MAWK->ps.program_string->len - 1] = '\n' ;
      MAWK->ps.buffp = (unsigned char *) MAWK->ps.program_string->str ;
      MAWK->ps.eof_flag = 1 ;
   }
   else	 /* program from file[s] */
   {
			/* loading the script can not fail yet, as this is the first file */
      mawk_scan_open(MAWK);
			if (MAWK->ps.buffer == NULL)
      	MAWK->ps.buffer = (unsigned char *) zmalloc(MAWK, BUFFSZ + 1) ;
			MAWK->ps.buffp = MAWK->ps.buffer;
			if (MAWK->do_exit)
				return;
     	scan_fillbuff(MAWK) ;
   }

#ifdef OS2  /* OS/2 "extproc" is similar to #! */
   if (strnicmp(MAWK->ps.buffp, "extproc ", 8) == 0)
     eat_comment(MAWK);
#endif
   mawk_eat_nl(MAWK) ;			 /* scan to first token */
   if (next(MAWK) == 0)
   {
      /* no program */
      mawk_exit(MAWK, 0) ;
   }

   un_next() ;

}

/* open MAWK->pfile_name[0]; returns 1 on success, 0 if the file could not
be open (error is also set) or is duplicate and ignored (normal condition) */
int
mawk_scan_open(mawk_state_t *MAWK)			/* open pfile_name */
{
   if (MAWK->ps.pfile_name[0] == '-' && MAWK->ps.pfile_name[1] == 0)
   {
      MAWK->ps.program_fd = 0 ;
   }
   else 
   {
			char *start, *end, *fn, *osp;
			int len, nlen, uniq;
			char path[PATH_MAX];
#ifdef mawk_realpath
			char rpath[PATH_MAX];
#else
#			define rpath path
#endif
			CELL idx;

			fn = MAWK->ps.pfile_name;
			if (*fn == '+') {
				fn++;
				uniq = 1;
			}
			else
				uniq = 0;

			nlen = strlen(fn);
			osp = ((STRING *)LIBPATH->ptr)->str;
			if ((osp == NULL) || (*osp == '\0') || (*fn == '/'))
				osp = "";

			for(end = start = osp; end != NULL; start = end+1)
   		{
				end = strchr(start, ';');
   			if (end == NULL)
   				len = strlen(start);
   			else
   				len = end-start;

   			if (len > 0) {
					memcpy(path, start, len);
					path[len] = '/';
					len++;
					path[len] = '\0';
				}
				else
					path[0] = '\0';

				if (len+nlen > sizeof(path)) {
		      mawk_errmsg(MAWK, errno, "cannot load awk script - path too long ('%s' '%s' vs %d bytes)", path, MAWK->ps.pfile_name, sizeof(path));
		      mawk_exit(MAWK, 2) ;
				}
				memcpy(path+len, fn, nlen);
				path[len+nlen] = '\0';

#ifdef mawk_realpath
				if (mawk_realpath(path, rpath) == NULL) {
		      mawk_errmsg(MAWK, errno, "cannot convert to realpath ('%s')", path);
		      mawk_exit(MAWK, 2) ;
				}
#endif

 				idx.type = C_STRING;
 				idx.ptr = mawk_new_STRING(MAWK, rpath);
				if (MAWK->scripts_loaded != NULL) {
					if (uniq) {
						/* we assume if an entry is in the array, the file exists,
						   as we already could load it once and there is no reason to assume
						   it has disappeared meanwhile - well, in theory this is a race
						   condition as another process may have deleted the file, but then
						   we would fail with "not found" anyway. */
						if (mawk_array_find(MAWK, MAWK->scripts_loaded, &idx, 0) != NULL)
							return 0;
					}
				}
				else
					MAWK->scripts_loaded = new_ARRAY();

   			if ((MAWK->ps.program_fd = open(path, O_RDONLY, 0)) != -1) {
					mawk_array_find(MAWK, MAWK->scripts_loaded, &idx, 1);
   				return 1;
   			}

   		}

      mawk_errmsg(MAWK, errno, "cannot open script %s (used search path %s)", fn, osp) ;
      mawk_exit(MAWK, 2) ;
   }
   return 0;
}

void
mawk_scan_cleanup(mawk_state_t *MAWK)
{
   if (MAWK->ps.program_fd >= 0)	{
   	zfree(MAWK, MAWK->ps.buffer, BUFFSZ + 1) ;
   	MAWK->ps.buffer = NULL;
   	MAWK->ps.buffp = NULL;
   }
   else	if (MAWK->ps.program_string != NULL)
   	free_STRING(MAWK->ps.program_string) ;

   if (MAWK->ps.program_fd > 0)	close(MAWK->ps.program_fd) ;

   /* redefine SPACE as [ \t\n] */

   MAWK->scan_code['\n'] = MAWK->posix_space_flag && MAWK->rs_shadow.type != SEP_MLR
      ? SC_UNEXPECTED : SC_SPACE ;
   MAWK->scan_code['\f'] = SC_UNEXPECTED ;	 /*value doesn't matter */
   MAWK->scan_code['\013'] = SC_UNEXPECTED ;	 /* \v not space */
   MAWK->scan_code['\r'] = SC_UNEXPECTED ;
}

/*----------------------------------------
 file reading functions
 next() and un_next(c) are macros in scan.h

 *---------------------*/

static void
scan_fillbuff(mawk_state_t *MAWK)
{
   unsigned r ;

   r = mawk_fillbuff(MAWK, MAWK->ps.program_fd, (char *) MAWK->ps.buffer, BUFFSZ, NULL) ;
   if (r < BUFFSZ)
   {
      MAWK->ps.eof_flag = 1 ;
      /* make sure eof is terminated */
      MAWK->ps.buffer[r] = '\n' ;
      MAWK->ps.buffer[r + 1] = 0 ;
   }
}

/* read one character -- slowly */
static int slow_next(mawk_state_t * MAWK)
{

	while (*MAWK->ps.buffp == 0) {

		if (!MAWK->ps.eof_flag) {
			MAWK->ps.buffp = MAWK->ps.buffer;
			scan_fillbuff(MAWK);
		}
		else {
			PFILE *q;

			if (MAWK->ps.program_fd > 0) {
				close(MAWK->ps.program_fd);
				MAWK->ps.program_fd = -1;
			}

			if (mawk_parser_pop(MAWK) == 0) {
				MAWK->ps.eof_flag = 0;
				do {
					if (MAWK->pfile_list != NULL) {
						MAWK->ps.pfile_name = MAWK->pfile_list->fname;
						q = MAWK->pfile_list;
						MAWK->pfile_list = MAWK->pfile_list->link;
						ZFREE(q);
					}
					else {
						MAWK->ps.eof_flag = 1;
						goto real_eof;
					}
				} while (mawk_scan_open(MAWK) != 1);
				MAWK->token_lineno = MAWK->lineno = 1;
			}
		}
	}

	real_eof:;
	return *MAWK->ps.buffp++;			/* note can un_next() , eof which is zero */
}

static void
eat_comment(mawk_state_t *MAWK)
{
   register int c ;

   while ((c = next(MAWK)) != '\n' && MAWK->scan_code[c]) ;
   un_next() ;
}

/* this is how we handle extra semi-colons that are
   now allowed to separate pattern-action blocks

   A proof that they are useless clutter to the language:
   we throw them away
*/

static void
eat_semi_colon(mawk_state_t *MAWK)
/* eat one semi-colon on the current line */
{
   register int c ;

   while (MAWK->scan_code[c = next(MAWK)] == SC_SPACE) ;
   if (c != ';')  un_next() ;
}

void
mawk_eat_nl(mawk_state_t *MAWK)			/* eat all space including newlines */
{
   while (1)
      switch (MAWK->scan_code[next(MAWK)])
      {
	 case SC_COMMENT:
	    eat_comment(MAWK) ;
	    break ;

	 case SC_NL:
	    MAWK->lineno++ ;
	    /* fall thru  */

	 case SC_SPACE:
	    break ;

	 case SC_ESCAPE:
	    /* bug fix - surprised anyone did this,
	       a csh user with backslash dyslexia.(Not a joke)
	    */
	    {
	       unsigned c ;

	       while (MAWK->scan_code[c = next(MAWK)] == SC_SPACE) ;
	       if (c == '\n')
		  MAWK->token_lineno = ++MAWK->lineno ;
	       else if (c == 0)  
	       {
		  un_next() ;
		  return ;
	       }
	       else /* error */
	       {
		  un_next() ;
		  /* can't un_next() twice so deal with it */
		  Mawk_lval.ival = '\\' ;
		  mawk_unexpected_char(MAWK) ;
		  if( ++(MAWK->compile_error_count) == MAX_COMPILE_ERRORS )
		     mawk_exit(MAWK, 2) ;
		  return ;
	       }
	    }
	    break ;
	     
	 default:
	    un_next() ;
	    return ;
      }
}

int
Mawk_lex(mawk_state_t *MAWK)
{
   register int c ;

	if (MAWK->do_exit)
		return -1;

   MAWK->token_lineno = MAWK->lineno ;

reswitch:

   switch (MAWK->scan_code[c = next(MAWK)])
   {
      case 0:
	 ct_ret(EOF) ;

      case SC_SPACE:
	 goto reswitch ;

      case SC_COMMENT:
	 eat_comment(MAWK) ;
	 goto reswitch ;

      case SC_NL:
	 MAWK->lineno++ ;
	 mawk_eat_nl(MAWK) ;
	 ct_ret(NL) ;

      case SC_ESCAPE:
	 while (MAWK->scan_code[c = next(MAWK)] == SC_SPACE) ;
	 if (c == '\n')
	 {
	    MAWK->token_lineno = ++MAWK->lineno ;
	    goto reswitch ;
	 }

	 if (c == 0)  ct_ret(EOF) ;
	 un_next() ;
	 Mawk_lval.ival = '\\' ;
	 ct_ret(UNEXPECTED) ;


      case SC_SEMI_COLON:
	 mawk_eat_nl(MAWK) ;
	 ct_ret(SEMI_COLON) ;

      case SC_LBRACE:
	 mawk_eat_nl(MAWK) ;
	 MAWK->brace_cnt++ ;
	 ct_ret(LBRACE) ;

      case SC_PLUS:
	 switch (next(MAWK))
	 {
	    case '+':
	       Mawk_lval.ival = '+' ;
	       string_buff[0] =
		  string_buff[1] = '+' ;
	       string_buff[2] = 0 ;
	       ct_ret(INC_or_DEC) ;

	    case '=':
	       ct_ret(ADD_ASG) ;

	    default:
	       un_next() ;
	       ct_ret(PLUS) ;
	 }

      case SC_MINUS:
	 switch (next(MAWK))
	 {
	    case '-':
	       Mawk_lval.ival = '-' ;
	       string_buff[0] =
		  string_buff[1] = '-' ;
	       string_buff[2] = 0 ;
	       ct_ret(INC_or_DEC) ;

	    case '=':
	       ct_ret(SUB_ASG) ;

	    default:
	       un_next() ;
	       ct_ret(MINUS) ;
	 }

      case SC_COMMA:
	 mawk_eat_nl(MAWK) ;
	 ct_ret(COMMA) ;

      case SC_MUL:
	 mawk_test1_ret('=', MUL_ASG, MUL) ;

      case SC_DIV:
	 {
	    static int can_precede_div[] =
	    {DOUBLE, STRING_, RPAREN, ID, D_ID, RE, RBOX, FIELD,
	     GETLINE, INC_or_DEC, -1} ;

	    int *p = can_precede_div ;

	    do
	    {
	       if (*p == MAWK->current_token)
	       {
		  if (*p != INC_or_DEC)	 mawk_test1_ret('=', DIV_ASG, DIV) ;

		  if (next(MAWK) == '=')
		  {
		     un_next() ;
		     ct_ret(collect_RE(MAWK)) ;
		  }
	       }
	    }
	    while (*++p != -1) ;

	    ct_ret(collect_RE(MAWK)) ;
	 }

      case SC_MOD:
	 mawk_test1_ret('=', MOD_ASG, MOD) ;

      case SC_POW:
	 mawk_test1_ret('=', POW_ASG, POW) ;

      case SC_LPAREN:
	 MAWK->paren_cnt++ ;
	 ct_ret(LPAREN) ;

      case SC_RPAREN:
	 if (--MAWK->paren_cnt < 0)
	 {
	    mawk_compile_error(MAWK, "extra ')'") ;
	    MAWK->paren_cnt = 0 ;
	    goto reswitch ;
	 }

	 ct_ret(RPAREN) ;

      case SC_LBOX:
	 ct_ret(LBOX) ;

      case SC_RBOX:
	 ct_ret(RBOX) ;

      case SC_MATCH:
	 string_buff[0] = '~' ;
	 string_buff[0] = 0 ;
	 Mawk_lval.ival = 1 ;
	 ct_ret(MATCH) ;

      case SC_EQUAL:
	 mawk_test1_ret('=', EQ, ASSIGN) ;

      case SC_NOT:		/* !  */
	 if ((c = next(MAWK)) == '~')
	 {
	    string_buff[0] = '!' ;
	    string_buff[1] = '~' ;
	    string_buff[2] = 0 ;
	    Mawk_lval.ival = 0 ;
	    ct_ret(MATCH) ;
	 }
	 else if (c == '=')  ct_ret(NEQ) ;

	 un_next() ;
	 ct_ret(NOT) ;


      case SC_LT:		/* '<' */
	 if (next(MAWK) == '=')  ct_ret(LTE) ;
	 else  un_next() ;

	 if (MAWK->getline_flag)
	 {
	    MAWK->getline_flag = 0 ;
	    ct_ret(IO_IN) ;
	 }
	 else  ct_ret(LT) ;

      case SC_GT:		/* '>' */
	 if (MAWK->print_flag && MAWK->paren_cnt == 0)
	 {
	    MAWK->print_flag = 0 ;
	    /* there are 3 types of IO_OUT
	       -- build the error string in string_buff */
	    string_buff[0] = '>' ;
	    if (next(MAWK) == '>')
	    {
	       Mawk_lval.ival = F_APPEND ;
	       string_buff[1] = '>' ;
	       string_buff[2] = 0 ;
	    }
	    else
	    {
	       un_next() ;
	       Mawk_lval.ival = F_TRUNC ;
	       string_buff[1] = 0 ;
	    }
	    return MAWK->current_token = IO_OUT ;
	 }

	 mawk_test1_ret('=', GTE, GT) ;

      case SC_OR:
	 if (next(MAWK) == '|')
	 {
	    mawk_eat_nl(MAWK) ;
	    ct_ret(OR) ;
	 }
	 else
	 {
	    un_next() ;

	    if (MAWK->print_flag && MAWK->paren_cnt == 0)
	    {
	       MAWK->print_flag = 0 ;
	       Mawk_lval.ival = PIPE_OUT ;
	       string_buff[0] = '|' ;
	       string_buff[1] = 0 ;
	       ct_ret(IO_OUT) ;
	    }
	    else  ct_ret(PIPE) ;
	 }

      case SC_AND:
	 if (next(MAWK) == '&')
	 {
	    mawk_eat_nl(MAWK) ;
	    ct_ret(AND) ;
	 }
	 else
	 {
	    un_next() ;
	    Mawk_lval.ival = '&' ;
	    ct_ret(UNEXPECTED) ;
	 }

      case SC_QMARK:
	 ct_ret(QMARK) ;

      case SC_COLON:
	 ct_ret(COLON) ;

      case SC_RBRACE:
	 if (--MAWK->brace_cnt < 0)
	 {
	    mawk_compile_error(MAWK, "extra '}'") ;
	    eat_semi_colon(MAWK) ;
	    MAWK->brace_cnt = 0 ;
	    goto reswitch ;
	 }

	 if ((c = MAWK->current_token) == NL || c == SEMI_COLON
	     || c == SC_FAKE_SEMI_COLON || c == RBRACE)
	 {
	    /* if the brace_cnt is zero , we've completed
	       a pattern action block. If the user insists
	       on adding a semi-colon on the same line
	       we will eat it.	Note what we do below:
	       physical law -- conservation of semi-colons */

	    if (MAWK->brace_cnt == 0)	 eat_semi_colon(MAWK) ;
	    mawk_eat_nl(MAWK) ;
	    ct_ret(RBRACE) ;
	 }

	 /* supply missing semi-colon to statement that
	     precedes a '}' */
	 MAWK->brace_cnt++ ;
	 un_next() ;
	 MAWK->current_token = SC_FAKE_SEMI_COLON ;
	 return SEMI_COLON ;

      case SC_DIGIT:
      case SC_DOT:
	 {
	    double d;
	    int flag ;
	    static double double_zero = 0.0 ;
	    static double double_one = 1.0 ;

	    if ((d = collect_decimal(MAWK, c, &flag)) == 0.0)
	    {
	       if (flag)  ct_ret(flag) ;
	       else  Mawk_lval.ptr = (PTR) & double_zero ;
	    }
	    else if (d == 1.0)
	    {
	       Mawk_lval.ptr = (PTR) & double_one ;
	    }
	    else
	    {
	       Mawk_lval.ptr = (PTR) ZMALLOC(double) ;
	       *(double *) Mawk_lval.ptr = d ;
	    }
	    ct_ret(DOUBLE) ;
	 }

      case SC_DOLLAR:		/* '$' */
	 {
	    double d;
	    int flag ;

	    while (MAWK->scan_code[c = next(MAWK)] == SC_SPACE) ;
	    if (MAWK->scan_code[c] != SC_DIGIT &&
		MAWK->scan_code[c] != SC_DOT)
	    {
	       un_next() ;
	       ct_ret(DOLLAR) ;
	    }

	    /* compute field address at compile time */
	    if ((d = collect_decimal(MAWK, c, &flag)) == 0.0)
	    {
	       if (flag)  ct_ret(flag) ; /* an error */
	       else  Mawk_lval.cp = &MAWK->field[0] ;
	    }
	    else
	    {
	       if (d > MAX_FIELD)
	       {
		  mawk_compile_error(
		     MAWK, "$%g exceeds maximum field(%d)", d, MAX_FIELD) ;
		  d = MAX_FIELD ;
	       }
	       Mawk_lval.cp = field_ptr((int) d) ;
	    }

	    ct_ret(FIELD) ;
	 }

      case SC_DQUOTE:
	 return MAWK->current_token = collect_string(MAWK) ;

      case SC_IDCHAR:		/* collect an identifier */
	 {
	    unsigned char *p =
	    (unsigned char *) string_buff + 1 ;
	    SYMTAB *stp ;

	    string_buff[0] = c ;

	    while (
		     (c = MAWK->scan_code[*p++ = next(MAWK)]) == SC_IDCHAR ||
		     c == SC_DIGIT) ;

	    un_next() ;
	    *--p = 0 ;

	    switch ((stp = mawk_find(MAWK, string_buff, 1))->type)
	    {
	       case ST_NONE:
		  /* check for function call before defined */
		  if (next(MAWK) == '(')
		  {
		     stp->type = ST_FUNCT ;
		     stp->stval.fbp = (FBLOCK *)
			zmalloc(MAWK, sizeof(FBLOCK)) ;
		     stp->stval.fbp->name = stp->name ;
		     stp->stval.fbp->code = (INST *) 0 ;
		     Mawk_lval.fbp = stp->stval.fbp ;
		     MAWK->current_token = FUNCT_ID ;
		  }
		  else
		  {
		     Mawk_lval.stp = stp ;
		     MAWK->current_token = MAWK->current_token == DOLLAR ? D_ID : ID ;
		  }
		  un_next() ;
		  break ;

	       case ST_NR:
		  MAWK->NR_flag = 1 ;
		  stp->type = ST_VAR ;
		  /* fall thru */

	       case ST_VAR:
	       case ST_ARRAY:
	       case ST_LOCAL_NONE:
	       case ST_LOCAL_VAR:
	       case ST_LOCAL_ARRAY:

		  Mawk_lval.stp = stp ;
		  MAWK->current_token = MAWK->current_token == DOLLAR ? D_ID : ID ;
		  break ;

	       case ST_ENV:
		  stp->type = ST_ARRAY ;
		  stp->stval.array = new_ARRAY() ;
		  mawk_load_environ(MAWK, stp->stval.array) ;
		  Mawk_lval.stp = stp ;
		  MAWK->current_token = MAWK->current_token == DOLLAR ? D_ID : ID ;
		  break ;

	       case ST_FUNCT:
		  Mawk_lval.fbp = stp->stval.fbp ;
		  MAWK->current_token = FUNCT_ID ;
		  break ;

	       case ST_C_FUNCTION:
		  Mawk_lval.fbp = calloc(sizeof(FBLOCK), 1);
		  Mawk_lval.fbp->name = strdup(string_buff);
		  Mawk_lval.fbp->code = NULL;
		  MAWK->current_token = C_FUNCT_ID ;
		  break ;


	       case ST_KEYWORD:
		  MAWK->current_token = stp->stval.kw ;
		  break ;

	       case ST_BUILTIN:
		  Mawk_lval.bip = stp->stval.bip ;
		  MAWK->current_token = BUILTIN ;
		  break ;

	       case ST_LENGTH:

		  Mawk_lval.bip = stp->stval.bip ;

		  /* check for length alone, this is an ugly
			 hack */
		  while (MAWK->scan_code[c = next(MAWK)] == SC_SPACE) ;
		  un_next() ;

		  MAWK->current_token = c == '(' ? BUILTIN : LENGTH ;
		  break ;

	       case ST_FIELD:
		  Mawk_lval.cp = stp->stval.cp ;
		  MAWK->current_token = FIELD ;
		  break ;


	       default:
	    fprintf(stderr, "stp type:%d\n", stp->type);
		  mawk_bozo(MAWK, "mawk_find returned bad st type") ;
	    }
	    return MAWK->current_token ;
	 }


      case SC_UNEXPECTED:
	 Mawk_lval.ival = c & 0xff ;
	 ct_ret(UNEXPECTED) ;
   }
   return 0 ;			 /* never get here make lint happy */
}

/* collect a decimal constant in temp_buff.
   Return the value and error conditions by reference */

static double
collect_decimal(MAWK, c, flag)
   mawk_state_t *MAWK;
   int c ;
   int *flag ;
{
   register unsigned char *p = (unsigned char *) string_buff + 1 ;
   unsigned char *endp ;
   double d;

   *flag = 0 ;
   string_buff[0] = c ;

   if (c == '.')
   {
      if (MAWK->scan_code[*p++ = next(MAWK)] != SC_DIGIT)
      {
	 *flag = UNEXPECTED ;
	 Mawk_lval.ival = '.' ;
	 return 0.0 ;
      }
   }
   else
   {
      while (MAWK->scan_code[*p++ = next(MAWK)] == SC_DIGIT) ;
      if (p[-1] != '.')
      {
	 un_next() ;
	 p-- ;
      }
   }
   /* get rest of digits after decimal point */
   while (MAWK->scan_code[*p++ = next(MAWK)] == SC_DIGIT) ;

   /* check for exponent */
   if (p[-1] != 'e' && p[-1] != 'E')
   {
      un_next() ;
      *--p = 0 ;
   }
   else	 /* get the exponent */
   {
      if (MAWK->scan_code[*p = next(MAWK)] != SC_DIGIT &&
	  *p != '-' && *p != '+')
      {
	 *++p = 0 ;
	 *flag = BAD_DECIMAL ;
	 return 0.0 ;
      }
      else  /* get the rest of the exponent */
      {
	 p++ ;
	 while (MAWK->scan_code[*p++ = next(MAWK)] == SC_DIGIT) ;
	 un_next() ;
	 *--p = 0 ;
      }
   }

   errno = 0 ;			 /* check for mawk_overflow/underflow */
   d = strtod(string_buff, (char **) &endp) ;

#ifndef	 STRTOD_UNDERFLOW_ON_ZERO_BUG
   if (errno)  mawk_compile_error(MAWK, "%s : decimal %sflow", string_buff,
		    d == 0.0 ? "under" : "over") ;
#else /* ! sun4 bug */
   if (errno && d != 0.0)
      mawk_compile_error(MAWK, "%s : decimal mawk_overflow", string_buff) ;
#endif

   if (endp < p)
   {
      *flag = BAD_DECIMAL ;
      return 0.0 ;
   }
   return d ;
}

/*----------  process escape characters ---------------*/

static const char hex_val['f' - 'A' + 1] = /* read-only */
{
   10, 11, 12, 13, 14, 15, 0, 0,
   0, 0, 0, 0, 0, 0, 0, 0,
   0, 0, 0, 0, 0, 0, 0, 0,
   0, 0, 0, 0, 0, 0, 0, 0,
   10, 11, 12, 13, 14, 15} ;

#define isoctal(x)  ((x)>='0'&&(x)<='7')

#define	 hex_value(x)	hex_val[(x)-'A']

#define ishex(x) (MAWK->scan_code[x] == SC_DIGIT ||\
		  'A' <= (x) && (x) <= 'f' && hex_value(x))

static int PROTO(octal, (char **)) ;
static int PROTO(hex, (mawk_state_t *, char **)) ;

/* process one , two or three octal digits
   moving a pointer forward by reference */
static int
octal(start_p)
   char **start_p ;
{
   register char *p = *start_p ;
   register unsigned x ;

   x = *p++ - '0' ;
   if (isoctal(*p))
   {
      x = (x << 3) + *p++ - '0' ;
      if (isoctal(*p))	x = (x << 3) + *p++ - '0' ;
   }
   *start_p = p ;
   return x & 0xff ;
}

/* process one or two hex digits
   moving a pointer forward by reference */

static int
hex(MAWK, start_p)
   mawk_state_t *MAWK;
   char **start_p ;
{
   register unsigned char *p = (unsigned char *) *start_p ;
   register unsigned x ;
   unsigned t ;

   if (MAWK->scan_code[*p] == SC_DIGIT)  x = *p++ - '0' ;
   else	 x = hex_value(*p++) ;

   if (MAWK->scan_code[*p] == SC_DIGIT)  x = (x << 4) + *p++ - '0' ;
   else if ('A' <= *p && *p <= 'f' && (t = hex_value(*p)))
   {
      x = (x << 4) + t ;
      p++ ;
   }

   *start_p = (char *) p ;
   return x ;
}

/* process the escape characters in a string, in place . */

char *
mawk_rm_escape(MAWK, s)
   mawk_state_t *MAWK;
   char *s ;
{
   register char *p, *q ;
   char *t ;
   int i ;

   q = p = s ;

   while (*p)
   {
      if (*p == '\\')
      {
	 MAWK->escape_test[ET_END].in = *++p ; /* sentinal */
	 i = 0 ;
	 while (MAWK->escape_test[i].in != *p)  i++ ;

	 if (i != ET_END)	/* in table */
	 {
	    p++ ;
	    *q++ = MAWK->escape_test[i].out ;
	 }
	 else if (isoctal(*p))
	 {
	    t = p ;
	    *q++ = octal(&t) ;
	    p = t ;
	 }
	 else if (*p == 'x' && ishex(*(unsigned char *) (p + 1)))
	 {
	    t = p + 1 ;
	    *q++ = hex(MAWK, &t) ;
	    p = t ;
	 }
	 else if (*p == 0)	/* can only happen with command line assign */
	    *q++ = '\\' ;
	 else  /* not an escape sequence */
	 {
	    *q++ = '\\' ;
	    *q++ = *p++ ;
	 }
      }
      else  *q++ = *p++ ;
   }

   *q = 0 ;
   return s ;
}

static int
collect_string(mawk_state_t *MAWK)
{
   register unsigned char *p = (unsigned char *) string_buff ;
   int c ;
   int e_flag = 0 ;		 /* on if have an escape char */

   while (1)
      switch (MAWK->scan_code[*p++ = next(MAWK)])
      {
	 case SC_DQUOTE:	/* done */
	    *--p = 0 ;
	    goto out ;

	 case SC_NL:
	    p[-1] = 0 ;
	    /* fall thru */

	 case 0:		/* unterminated string */
	    mawk_compile_error(
			    MAWK, "runaway string constant \"%.10s ...",
			    string_buff, MAWK->token_lineno) ;
	    mawk_exitval(MAWK, 2, -1) ;

	 case SC_ESCAPE:
	    if ((c = next(MAWK)) == '\n')
	    {
	       p-- ;
	       MAWK->lineno++ ;
	    }
	    else if (c == 0)  un_next() ;
	    else
	    {
	       *p++ = c ;
	       e_flag = 1 ;
	    }

	    break ;

	 default:
	    break ;
      }

out:
   Mawk_lval.ptr = (PTR) mawk_new_STRING(MAWK,
				   e_flag ? mawk_rm_escape(MAWK, string_buff)
				   : string_buff) ;
   return STRING_ ;
}


static int
collect_RE(mawk_state_t *MAWK)
{
   register unsigned char *p = (unsigned char *) string_buff ;
   int c ;
   STRING *sval ;

   while (1)
      switch (MAWK->scan_code[*p++ = next(MAWK)])
      {
	 case SC_DIV:		/* done */
	    *--p = 0 ;
	    goto out ;

	 case SC_NL:
	    p[-1] = 0 ;
	    /* fall thru */

	 case 0:		/* unterminated re */
	    mawk_compile_error(
			    MAWK, "runaway regular expression /%.10s ...",
			    string_buff, MAWK->token_lineno) ;
	    mawk_exitval(MAWK, 2, -1) ;

	 case SC_ESCAPE:
	    switch (c = next(MAWK))
	    {
	       case '/':
		  p[-1] = '/' ;
		  break ;

	       case '\n':
		  p-- ;
		  break ;

	       case 0:
		  un_next() ;
		  break ;

	       default:
		  *p++ = c ;
		  break ;
	    }
	    break ;
      }

out:
   /* now we've got the RE, so compile it */
   sval = mawk_new_STRING(MAWK, string_buff) ;
   Mawk_lval.ptr = mawk_re_compile(MAWK, sval) ;
   free_STRING(sval) ;
   return RE ;
}

void mawk_parser_push(mawk_state_t *MAWK)
{
	if (MAWK->pstack_used >= MAWK->pstack_alloced) {
		MAWK->pstack_alloced += 8;
		MAWK->mawk_parser_stack = realloc(MAWK->mawk_parser_stack, sizeof(mawk_parse_state_t) * MAWK->pstack_alloced);
	}
	memcpy(&MAWK->mawk_parser_stack[MAWK->pstack_used], &MAWK->ps, sizeof(mawk_parse_state_t));
	MAWK->pstack_used++;
}

int mawk_parser_pop(mawk_state_t *MAWK)
{
	if (MAWK->pstack_used <= 0)
		return 0;

	zfree(MAWK, MAWK->ps.buffer, BUFFSZ+1);
	MAWK->ps.buffer = NULL;
	MAWK->ps.buffp  = NULL;

	MAWK->pstack_used--;
	memcpy(&MAWK->ps, &MAWK->mawk_parser_stack[MAWK->pstack_used], sizeof(mawk_parse_state_t));

	if (MAWK->pstack_used == 0) {
		free(MAWK->mawk_parser_stack);
		MAWK->mawk_parser_stack = NULL;
	}

	return 1;
}
